use kahlo::math::{PixelBox, PixelPoint, PixelSize};
use std::ops::Add;

use super::{cache::NodeState, LayoutNodeAccess, SizePolicy};

pub trait ArrangementCalculator {
    fn arrange_step(
        &self,
        node: LayoutNodeAccess,
        child_policies: Vec<(SizePolicy, SizePolicy)>,
    ) -> (SizePolicy, SizePolicy);
    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox);
}

#[derive(Clone, Debug)]
pub enum LineArrangement {
    Column,
    Row,
}

impl LineArrangement {
    fn is_column(&self) -> bool {
        match self {
            Self::Column => true,
            Self::Row => false,
        }
    }
}

impl ArrangementCalculator for LineArrangement {
    fn arrange_step(
        &self,
        node: LayoutNodeAccess,
        child_policies: Vec<(SizePolicy, SizePolicy)>,
    ) -> (SizePolicy, SizePolicy) {
        if child_policies.is_empty() {
            return (node.width_policy, node.height_policy);
        }

        let cw_policy = child_policies
            .iter()
            .map(|v| v.0)
            .reduce(if self.is_column() {
                SizePolicy::max
            } else {
                SizePolicy::add
            })
            .unwrap();
        let ch_policy: SizePolicy = child_policies
            .iter()
            .map(|v| v.1)
            .reduce(if self.is_column() {
                SizePolicy::add
            } else {
                SizePolicy::max
            })
            .unwrap();
        (
            node.width_policy.max_preserve_slack(cw_policy),
            node.height_policy.max_preserve_slack(ch_policy),
        )
    }

    fn layout_step(&self, node: LayoutNodeAccess, inside: PixelBox) {
        // do the final children layouts
        node.cache
            .with_state(node.cache_key, |ns| {
                if Some(inside) != ns.area {
                    ns.area = Some(inside);
                    ns.needs_render = true;
                }
            });

        if node.child_len() == 0 {
            return;
        }

        // expansion direction extraction lambda
        let ee = if self.is_column() {
            |ns: &mut NodeState| ns.net_policy.1
        } else {
            |ns: &mut NodeState| ns.net_policy.0
        };

        let policies = node
            .child_iter()
            .map(|c| node.cache.with_state(c.cache_key, ee).unwrap());
        let mut last_offset = 0;
        let fit = do_fit(
            if self.is_column() {
                inside.height() as i32
            } else {
                inside.width() as i32
            },
            policies,
        );

        for (offset, child) in fit.zip(node.child_iter()) {
            let cbox = if self.is_column() {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x,
                        y: inside.min.y + last_offset,
                        ..Default::default()
                    },
                    PixelSize {
                        width: inside.width(),
                        height: (offset - last_offset) as i32,
                        ..Default::default()
                    },
                )
            } else {
                PixelBox::from_origin_and_size(
                    PixelPoint {
                        x: inside.min.x + last_offset,
                        y: inside.min.y,
                        ..Default::default()
                    },
                    PixelSize {
                        width: (offset - last_offset) as i32,
                        height: inside.height(),
                        ..Default::default()
                    },
                )
            };

            self.layout_step(child, cbox);
            last_offset = offset;
        }
    }
}

/// Calculate the fit of a number of size policies inside a given length.
///
/// Returns the endpoints of each item.
fn do_fit<'l>(
    total: i32,
    policies: impl Iterator<Item = SizePolicy> + Clone,
) -> impl Iterator<Item = i32> {
    // first pass over children: collect sum total of minimum/desired sizes and/slack weights
    let policy_sum = policies.clone().reduce(SizePolicy::add).unwrap_or_default();
    let (sum_min, sum_desired, sum_slack) = (policy_sum.minimum as i32, policy_sum.desired as i32, policy_sum.slack_weight as i32);

    if total < sum_min {
        todo!("not enough space to distribute all children")
    }

    let (fit_space, slack_space) = if sum_desired <= total {
        (sum_desired - sum_min, total - sum_desired)
    } else {
        (total - sum_min, 0)
    };

    let total_fit_units = sum_desired - sum_min;

    let (fit_per, fit_per_rem) = if total_fit_units > 0 {
        (fit_space / total_fit_units, fit_space % total_fit_units)
    } else {
        (0, 0)
    };

    let (slack_per, slack_per_rem) = if sum_slack > 0 {
        (slack_space / sum_slack, slack_space % sum_slack)
    } else {
        (0, 0)
    };

    let mut allocated_fit = 0;
    let mut allocated_fit_rem = 0;

    let mut allocated_slack = 0;
    let mut allocated_slack_rem = 0;

    let mut min_allocation = 0;

    policies.map(move |w| {
        min_allocation += w.minimum as i32;

        let fit_units = (w.desired - w.minimum) as i32;
        if fit_units > 0 {
            allocated_fit += fit_per * fit_units;
            allocated_fit_rem += fit_per_rem * fit_units;
            allocated_fit += allocated_fit_rem / total_fit_units;
            allocated_fit_rem %= total_fit_units;
        }

        let slack_units = w.slack_weight as i32;
        if slack_units > 0 {
            allocated_slack += slack_per * slack_units;
            allocated_slack_rem += slack_per_rem * slack_units;
            allocated_slack += allocated_slack_rem / sum_slack;
            allocated_slack_rem %= sum_slack;
        }

        min_allocation + allocated_fit + allocated_slack
    })
}

#[cfg(test)]
mod tests {
    use crate::layout::SizePolicy;

    use super::do_fit;

    #[test]
    fn distribute_2_even() {
        let policies = vec![SizePolicy::expanding(1), SizePolicy::expanding(1)];

        assert_eq!(
            do_fit(2, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2]
        );
        assert_eq!(
            do_fit(3, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 3]
        );
    }

    #[test]
    fn distribute_3_even() {
        let policies = vec![SizePolicy::expanding(1), SizePolicy::expanding(1), SizePolicy::expanding(1)];

        assert_eq!(
            do_fit(3, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2, 3]
        );
        assert_eq!(
            do_fit(4, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 2, 4]
        );
        assert_eq!(
            do_fit(5, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![1, 3, 5]
        );
        assert_eq!(
            do_fit(6, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![2, 4, 6]
        );
        assert_eq!(
            do_fit(7, policies.clone().into_iter()).collect::<Vec<_>>(),
            vec![2, 4, 7]
        );
    }
}
