use std::cell::RefCell;

use kahlo::{math::{PixelBox, PixelSideOffsets}, prelude::*};

use crate::{
    component::Component,
    layout::{LayoutNode, LayoutNodeAccess, LeafLayoutNode, SizePolicy, HorizontalAlignment},
    theme::Theme,
    ui::UIHandle, window::RenderFormat,
};

use super::Widget;

pub struct Label<C: Component> {
    lnode: LeafLayoutNode,
    text: Option<String>,
    rendered: RefCell<Option<kahlo::Alphamap>>,
    _ghost: std::marker::PhantomData<C>,
}

impl<C: Component> Label<C> {
    pub fn new(uih: &UIHandle) -> Self {
        let mut node = uih.new_layout_node();
        node.set_width_policy(SizePolicy::expanding(1))
            .set_height_policy(SizePolicy::fixed(uih.theme().ui_font_size.ceil() as usize));
        Self {
            lnode: LeafLayoutNode::new(node),
            text: None,
            rendered: None.into(),
            _ghost: Default::default(),
        }
    }

    pub fn set_text(&mut self, text: &str) {
        if Some(text) == self.text.as_deref() {
            return;
        } else {
            self.text = Some(text.to_string());
            self.rendered.take();
        }
    }
}

impl<C: Component> Label<C> {
    fn update_hints(&mut self, uih: &UIHandle) {
        let Some(text) = self.text.as_ref() else {
            return;
        };
        let line = uih.theme().make_line(text.as_str());
        let rendered = line.render_line();
        *self.rendered.borrow_mut() = Some(rendered);
        self.lnode.render_needed();
    }
}

impl<C: Component> Widget<C> for Label<C> {
    fn poll(
        &mut self,
        uih: &mut UIHandle,
        _input_state: Option<&crate::input::InputState>,
    ) -> Vec<<C as Component>::Msg> {
        if self.rendered.borrow().is_none() {
            self.update_hints(uih);
        }

        vec![]
    }

    fn layout_node(&self) -> LayoutNodeAccess {
        LayoutNodeAccess::new(&self.lnode)
    }
    fn layout_node_mut(&mut self) -> &mut LayoutNode {
        &mut self.lnode
    }

    fn render(&self, theme: &Theme, surface: &mut kahlo::BitmapMut<RenderFormat>) {
        if !self.lnode.render_check() {
            return
        }
        let rcon = self.rendered.borrow();
        let Some(rendered) = rcon.as_ref() else {
            return;
        };
        let render_location = self.layout_node().render_area().unwrap();

        // perform horizontal alignment
        let alignment_slack = (render_location.width() - rendered.size().width).max(0);
        let render_location = match self.lnode.halign() {
            HorizontalAlignment::Left => render_location.inner_box(PixelSideOffsets::new(0, alignment_slack, 0, 0)),
            HorizontalAlignment::Centre => render_location.inner_box(PixelSideOffsets::new(0, alignment_slack / 2, 0, alignment_slack / 2)),
            HorizontalAlignment::Right => render_location.inner_box(PixelSideOffsets::new(0, 0, 0, alignment_slack)),
        };

        surface.fill_region_masked(
            &rendered,
            PixelBox::from_size(rendered.size()),
            render_location.min,
            theme.foreground,
            kahlo::colour::BlendMode::SourceAlpha,
        );
    }
}
